import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc

""" The subroutine gammat corresponds to the transpose of gamma 
referenced in the text.  One action of gammat is to take the continuous 
data and create the combed data, that is, the pseudo-deblended data.
"""

def gammat(sdata,randtimes, ntimes,ntimespad,nxtrc,nxtrcpad,nytrc,nytrcpad):

	""" 
	Transpose gamma, make combed data from the continuos. 
	
	sdata is the continuous data,
	randtimes is the array of random source times in samples,
	ntimes is the length of the source traces to be generated,
	ntimespad is the trace length padded for the fft,
	nxtrc is the number of traces in the x-direction of the receiver gather,
	nxtrcpad is the number of traces in the x-direction of
	the receiver gather padded for the fft,
	nytrc is the number of traces in the y-direction of the receiver gather,
	nytrcpad is the number of traces in the y-direction of
	the receiver gather padded for the fft,	
	"""

	mpad = np.zeros([nxtrcpad,nytrcpad,ntimespad])

	for iy in range(nytrc):
		for ix in range(nxtrc):
			ipoint = int(randtimes[ix][iy])-1
			mpad[ix,iy,0:ntimes] = sdata[ipoint:ipoint+ntimes]

	return mpad

"""
The subroutine gammaf is the forward action of gamma referenced in the text.
One action of gammaf is to build the continuous data from the non-simultaneous
source data.  
"""

def gammaf(inarray,randtimes,ntimes,nxtrc,nytrc):

	""" 
	Forward gamma, make continuous data from the model. 
 	inarray is the input buffer for the separated sources. 
 	randtimes is the array of random source times in samples,
 	ntimes is the length of the source traces to be generated,
	nxtrc is the number of traces in the x-direction of the receiver gather,
	nytrc is the number of traces in the y-direction of the receiver gather.
	
	The output is the continuos record generated from the input receiver 
	gather.
	"""

	lengthdata = np.ceil( np.max(randtimes) + ntimes )
	sdata = np.zeros(int(lengthdata))
#
	for iy in range(nytrc):
		for ix in range(nxtrc):
			ipoint = int(randtimes[ix][iy])-1
			sdata[ipoint:ipoint+ntimes] = \
			inarray[ix,iy,0:ntimes]+\
			sdata[ipoint:ipoint+ntimes] 


	return sdata

"""
This sparse inversion subroutine uses a greedy algorithm to solve for 
the separated seismic data, that is, the deblended data.  
"""

def sparseinverse(sdata,ntimespad,nxtrcpad,nytrcpad,\
                  ntimes,nxtrc,nytrc, \
                  niter,maxblendfact,randtimes):

	""" 
	sdata is the input continuous data,
	ntimespad is the trace length padded for the fft,
	nxtrcpad is the number of traces in the x-direction of 
	the receiver gather padded for the fft,
	nytrcpad is the number of traces in the y-direction of 
	the receiver gather padded for the fft,
	ntimes is is the length of the source traces extraced from 
	the continuous data,
	nxtrc is the number of traces in the x-direction of the receiver gather,
	nytrc is the number of traces in the y-direction of the receiver gather.
	niter is the number of iterations to use in the solver,
	maxblendfact is the maximum blending factor, used for stability,
	randtimes is the array of random source times in samples.

	The output mpad is a zero-padded receiver gather.

	"""

	mpad = np.zeros([nxtrcpad,nytrcpad,ntimespad])

	combed = gammat(sdata,randtimes,ntimes,ntimespad,\
	nxtrc,nxtrcpad,nytrc,nytrcpad)

	fcombed = np.fft.fftn(combed)
	maxcoef = np.max(np.abs(fcombed))  ## thresholding limit
	lengthdata = np.max(randtimes) + ntimes
	dp = np.zeros(int(lengthdata))

	for iter in range(1,niter+1):
		dnew = sdata - dp
		dx = gammat(dnew,randtimes,ntimes,ntimespad,\
		            nxtrc,nxtrcpad,nytrc,nytrcpad)
		mshotsv=np.copy(mpad)
		mpad = mpad + dx

		freqdata = np.fft.fftn(mpad)
		tolerance = maxcoef * ( ( float(niter) - iter )/niter)**2

		ir = np.shape(mpad)[0]
		jr = np.shape(mpad)[1]
		kr = np.shape(mpad)[2]

		# zero everything below tolerance
		freqdata=freqdata*np.greater(np.absolute(freqdata),tolerance)

		mpad = np.real(np.fft.ifftn(freqdata))
		temp = mpad - mshotsv
		mpad = mshotsv + temp * 0.72 / maxblendfact #  stabilize

		dp = gammaf(mpad,randtimes,ntimes,nxtrc,nytrc); #calculated data

	return mpad



# File: deblend.py. Separation quality diagnostic

""" Module deblend: This routine sets up a simple synthetic receiver gather 
and creates the corresponding continuous data.  It then deblends the 
continuous data to obtain the deblended data.

"""

def deblend(niter):

	""" 
	Separate the overlapping sources.  
	niter is the number of iterations to use to solve the separation.
	"""


	ntimes = 201  # number of time samples to generate
	nxtrc = 30    # number of traces in the x-dimension
	nytrc = 5    # number of traces in the y-dimension
	dt = 0.004    # sample interval

	# padded arrays
	ntimespad = ntimes * 2
	nxtrcpad = nxtrc * 2 
	nytrcpad = nytrc * 2 

	m = np.zeros([nxtrc,nytrc,ntimes])
	mpad = np.zeros([nxtrcpad,nytrcpad,ntimespad])

	m[:,:,40-1] = 1000.0
	m[:,:,60-1] = 1000.0
	m[:,:,80-1] = 1000.0
	m[:,:,100-1] = 1000.0
	m[:,:,120-1] = 1000.0
	m[:,:,140-1] = 1000.0
	m[:,:,160-1] = 1000.0
	m[:,:,180-1] = 1000.0
	m[:,:,200-1] = 1000.0

# - plot original m ---
	rc('font',family='serif')

	fig = plt.figure()
	mmm = np.reshape(m,(nxtrc*nytrc,ntimes))
	plt.imshow(np.transpose(mmm),aspect='auto',cmap='gray')
	plt.title('m - unpadded input')
	ttt = 'unpadded input, niter= ' + str(niter)
	plt.title(ttt)
	plt.colorbar()	

# set up the synthetic data

	tau = round(0.250/dt)   # amount of randomness, +-250 ms is the maximum
	randtimes = (np.random.rand(nxtrc,nytrc)-0.5)*2
	randtimes = np.rint(randtimes * tau + ntimes/2.0 )
	randtimes = np.cumsum(randtimes)
	randtimes = np.reshape(randtimes,(nxtrc,nytrc))

	onearray = np.ones_like(m)
	sdata = gammaf(onearray,randtimes, ntimes,nxtrc,nytrc)
	maxblendfact = max(sdata)
	print( ' maximum blend factor =',maxblendfact,"\n" )

	sdata = gammaf(m,randtimes, ntimes,nxtrc,nytrc)
	combed = gammat(sdata,randtimes,ntimes,ntimespad,\
	nxtrc,nxtrcpad,nytrc,nytrcpad)


#     solve for the separated sources 
	mpad = sparseinverse(sdata,ntimespad,nxtrcpad,nytrcpad,\
	ntimes,nxtrc,nytrc,\
	niter,maxblendfact,randtimes)

#  prepare results for display
	newm = np.zeros([nxtrc,nytrc,ntimes])
	newcombed = np.zeros([nxtrc,nytrc,ntimes])
	ir = np.shape(newm)[0]
	jr = np.shape(newm)[1]
	kr = np.shape(newm)[2]
	for i in range(ir):
		for j in range(jr):
			for k in range(kr):
				newm[i][j][k] = mpad[i][j][k] 
				newcombed[i][j][k] = combed[i][j][k] 

	diff = m - newm 

#  display calculated data

	rc('font',family='serif')
	fig = plt.figure()
	mmm = np.reshape(newm,(nxtrc*nytrc,ntimes))
	plt.imshow(np.transpose(mmm),aspect='auto',cmap='gray')
	ttt = 'calculated data, niter= ' + str(niter)
	plt.title(ttt)
	plt.colorbar()	

#  display combed data

	rc('font',family='serif')
	fig = plt.figure()
	mmm = np.reshape(newcombed,(nxtrc*nytrc,ntimes))
	plt.imshow(np.transpose(mmm),aspect='auto',cmap='gray')
	ttt = 'combed data, niter= ' + str(niter)
	plt.title(ttt)
	plt.colorbar()	

#  display diff data

	rc('font',family='serif')
	fig = plt.figure()
	mmm = np.reshape(diff,(nxtrc*nytrc,ntimes))
	plt.imshow(np.transpose(mmm),aspect='auto',cmap='gray')
	ttt = 'difference output, niter= ' + str(niter)
	plt.title(ttt)
	plt.colorbar()	

	# uncomment the next two lines if you want to save a plot
	#file = 'diff_niter_'  + str(niter) + '.tiff'
	#plt.savefig(file)
	
	plt.show()

	return 



#################################\

niter = 1000
print (' number of iterations =', niter)
deblend(niter)

